

set -ex



pip check
echo "*** Test Cases ***"                   >> test-6.1.1.robot
echo "Hello World"                          >> test-6.1.1.robot
echo "    Log   %{PKG_NAME} %{PKG_VERSION}" >> test-6.1.1.robot
robot test-6.1.1.robot
rebot output.xml
libdoc test-6.1.1.robot test.html
exit 0
